#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import re
from nltk.tokenize import word_tokenize, sent_tokenize
import spacy

text = text = re.sub(r'(\([^()]*?)(\.)([^()]*?\))', r'\1__.__\3', "fjdsk fds __CODE__. fdjslk")
print(text)
print(sent_tokenize(text))
print(word_tokenize("The system's behavior after you declare a permission depends on how sensitive the permission is org.appache.hadoop.Add().add(int,int)."))
        # f_out.write(sent + "\n")
        # f_out.write(nltk_tokens + "\n")
        # f_out.write(spacy_tokens + "\n")
        # f_out.write("\n")

# from kgtools.wrapper import Lazy

# class Test:
#     def __init__(self, text):
#         self.text = text

#     @Lazy
#     def lemma(self):
#         return self.text + "lemma"

# t = Test("test")
# print(t.lemma)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kgtools.nlp.tokenizer import CompoundTokenizer

tokenizer = CompoundTokenizer()
sent = tokenizer.word_tokenize("The word-frequency returns the sequence element's frequency over training corpus.")
print(sent)
for token in sent:
    print(token)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import threading
import multiprocessing
import random
from functools import wraps
from pathos.multiprocessing import ProcessingPool as Pool

from kgtools.func import reduce_seqs

WORKERS = multiprocessing.cpu_count() - 1


def TimeLog(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        print("@TimeLog[%s]: %s starts." % (time.strftime("%X", time.localtime()), fn.__qualname__))
        start = time.time()
        rs = fn(*args, **kwargs)
        end = time.time()
        print("@TimeLog[%s]: %s finishes." % (time.strftime("%X", time.localtime()), fn.__qualname__))
        print("@TimeLog[%fs]: %s takes." % (end - start, fn.__qualname__))
        return rs
    return wrapper


class Lazy(object):
    def __init__(self, func):
        self.func = func
        print("@Lazy[%s]: lazy property is declared." % self.func.__qualname__)

    def __get__(self, instance, cls):
        val = self.func(instance)
        setattr(instance, self.func.__name__, val)
        return val


def ID(*prime_keys):
    def wrapper(clazz):
        clazz.__str__ = lambda obj: '<{}: {}>'.format(obj.__class__.__name__, " ".join([f"{key}={str(obj.__dict__[key])}" for key in prime_keys]))
        clazz.__repr__ = lambda obj: '<{}: {}>'.format(obj.__class__.__name__, " ".join([f"{k}={str(v)}" for k, v in obj.__dict__.items()]))
        clazz.__eq__ = lambda obj, other: hash(obj) == hash(other)
        clazz.__hash__ = lambda obj: hash(str(obj))
        return clazz
    return wrapper


def Singleton(clazz):
    def new(cls, *args, **kwargs):
        if not hasattr(cls, "__instance"):
            with threading.Lock():
                if not hasattr(cls, "__instance"):
                    print("@Singleton[%s]: initialize singleton object." % cls.__name__)
                    cls.__instance = object.__new__(cls)
                    cls.__instance.__init__(*args, **kwargs)
        # print("@Singleton[%s]: get singleton object." % cls.__name__)
        return cls.__instance

    if not hasattr(clazz, "__Singleton"):
        with threading.Lock():
            if not hasattr(clazz, "__Singleton"):
                clazz.__Singleton = True
                clazz.__new__ = new
    return clazz


def Cache(fn):
    cache = {}
    print("@Cache[%s]: add cache." % fn.__qualname__)

    @wraps(fn)
    def wrapper(*arg):
        if arg not in cache:
            result = fn(*arg)
            cache[arg] = result
        return cache[arg]
    return wrapper


def Parallel(workers=WORKERS, batch_size=1000, shuffle=True, after_hook=None):
    def outer(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            print(f"@Parallel[workers={workers}, batch_size={batch_size}]: parallel for {fn.__qualname__}.")

            obj, data, _args = tuple(), tuple(), tuple()
            if hasattr(args[0].__class__, fn.__name__):
                obj, data, *_args = args
                obj = (obj, )
            else:
                data, *_args = args

            if type(data) != list:
                data = list(data)

            # assert type(data) == list, "Type of data must be list"

            if shuffle:
                print(f"@Parallel[workers={workers}, batch_size={batch_size}]: shuffle data for {fn.__qualname__}.")
                random.shuffle(data)

            pool = Pool(workers)
            pool.terminate()
            pool.restart()

            total_size = len(data)

            proc = []
            for beg, end in zip(range(0, total_size, batch_size), range(batch_size, total_size + batch_size, batch_size)):
                batch = data[beg:end]
                p = pool.apipe(fn, *obj, batch, *_args, **kwargs)
                proc.append(p)
            pool.close()
            pool.join()

            result = reduce_seqs([p.get() for p in proc])
            if after_hook is not None:
                result = after_hook(result)

            return result
        return wrapper
    return outer


if __name__ == "__main__":
    # import re

    # class Decamelizer:
    #     def __init__(self):
    #         pass

    #     @Cache
    #     def foo(self, camel_case):
    #         decamelized = re.sub(r'([A-Za-z])(2|4)([A-CE-Za-ce-z])', r'\1 \2 \3', camel_case).strip()
    #         decamelized = re.sub(r'_', " ", decamelized)
    #         decamelized = re.sub(r'([A-Z]+)([A-Z][a-z0-9]+)', r'\1 \2', decamelized)
    #         decamelized = re.sub(r'([0-9]?[A-Z]+)', r' \1', decamelized)
    #         decamelized = re.sub(r'\s+', " ", decamelized).strip()
    #         return decamelized

    #     def bar(self, camel_case):
    #         decamelized = re.sub(r'([A-Za-z])(2|4)([A-CE-Za-ce-z])', r'\1 \2 \3', camel_case).strip()
    #         decamelized = re.sub(r'_', " ", decamelized)
    #         decamelized = re.sub(r'([A-Z]+)([A-Z][a-z0-9]+)', r'\1 \2', decamelized)
    #         decamelized = re.sub(r'([0-9]?[A-Z]+)', r' \1', decamelized)
    #         decamelized = re.sub(r'\s+', " ", decamelized).strip()
    #         return decamelized

    #     def __call__(self, camel_case):
    #         return self.foo(camel_case)

    # decamelizer = Decamelizer()
    # import time
    # start = time.time()
    # for _ in range(10000):
    #     decamelizer("StringBuilder")
    # end = time.time()
    # print(end - start)

    # start = time.time()
    # for _ in range(10000):
    #     decamelizer.bar("StringBuilder")
    # end = time.time()
    # print(end - start)
    # class Foo:
    #     def __init__(self):
    #         pass

    #     @Parallel(clazz="Foo")
    #     def split(self, data):
    #         return [s.split() for s in data]

    # @Parallel(clazz=None)
    # def foo(data):
    #     return [s.split() for s in data]

    # rs = foo(["a b c"] * 100)
    # print(rs)

    @Singleton
    class Foo:
        def __init__(self, a=1):
            self.a = a
            self.list = []

    foo1 = Foo()
    foo1.a = 2
    foo1.list.append(1)
    foo2 = Foo()
    print(foo1 == foo2)
    print(foo2.a)
    print(foo2.list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import dill
import json
from pathlib import Path
from enum import Enum, unique


@unique
class FileFormat(Enum):
    JSON = "json"
    BINARY = "binary"


class Saver:
    @staticmethod
    def load(file_name, file_format=FileFormat.BINARY):
        if file_format == FileFormat.BINARY:
            with Path(file_name).open("rb") as f:
                return dill.load(f)
        elif file_format == FileFormat.JSON:
            with Path(file_name).open("r", encoding="utf-8") as f:
                return json.load(f)
        else:
            raise NotImplementedError

    @staticmethod
    def dump(obj, file_name, file_format=FileFormat.BINARY):
        Path(file_name).parent.mkdir(parents=True, exist_ok=True)
        if file_format == FileFormat.BINARY:
            with Path(file_name).open("wb") as f:
                dill.dump(obj, f)
        elif file_format == FileFormat.JSON:
            with Path(file_name).open("w", encoding="utf-8") as f:
                json.dump(obj, f, indent=4)
        else:
            raise NotImplementedError

    @staticmethod
    def transform(file, src_fmt, tgt_fmt):
        Saver.dump(Saver.load(file, src_fmt), tgt_fmt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import asyncio
import aiohttp
import async_timeout
import re
from bs4 import BeautifulSoup

from kgtools.func import reduce_seqs
from kgtools.saver import Saver, FileFormat
from kgtools.annotation import TimeLog


class Spider:

    def __init__(self, root, upper=None, lower=None, proxy_server=None, pool_size=63, retry=3):
        self.root = root
        simple_root = re.sub(r'(http://|https://)?(.*)', r'\2', root)
        self.domain = re.sub(r'(http://|https://)?(.*)', r'\1', root) + simple_root.split("/")[0]
        self.upper = upper if upper is not None else self.domain
        self.lower = lower if lower is not None else "<NO-LOWER>"

        self.storage = {}
        self.waiting_urls = set() if root is None else {root}
        self.fialed_urls = set()
        self.retry = retry

        self.proxy_server = proxy_server
        self.semaphore = asyncio.Semaphore(pool_size)

    def __is_in_scope(self, url):
        if url.startswith(self.upper) and not url.startswith(self.lower):
            return True
        else:
            return False

    def __normalize_url(self, cur_url, href):
        # cur_url = re.sub(r'(http://|https://)?(.*)', r'\2', cur_url)
        if len(href) == 0 or href.startswith("#") or href.startswith("?"):
            return cur_url
        if href.startswith("http://") or href.startswith("https://"):
            url = href
        elif href.startswith("/") and len(href) > 1:
            url = self.domain.rstrip("/") + "/" + href.lstrip("/")
        else:
            url = "/".join(cur_url.split("/")[:-1]).rstrip("/") + "/" + href.lstrip("/")
        url = re.sub(r'([^#\?]*)(#.*|\?.*)', r'\1', url)
        parts = []
        eles = url.split("/")
        for ele in eles:
            if ele == ".":
                continue
            if ele == "..":
                parts.pop()
                continue
            parts.append(ele)
        url = "/".join(parts)
        return url

    async def __request(self, url):
        try:
            async with self.semaphore:
                async with aiohttp.ClientSession() as session:
                    with async_timeout.timeout(10):
                        async with session.get(url, proxy=self.proxy_server) as response:
                            return await response.text()
        except Exception:
            print(f"[Failed] {url}")
            self.fialed_urls.add(url)
            return None

    async def fetch(self, url, recursive=True):
        links = set()
        if url in self.storage:
            return links
        html = await self.__request(url)
        if html is None:
            return links
        body = BeautifulSoup(html, "lxml").body
        if body is None:
            return links
        self.storage[url] = str(body)
        if recursive:
            for a in body.findAll("a"):
                link = a.get("href", "")
                if len(link) > 0:
                    link = self.__normalize_url(url, link)
                    if self.__is_in_scope(link):
                        links.add(link)
        print(f"[Done] {url}")
        return links

    @TimeLog
    def start_crawl(self, recursive_depth=None):
        current_depth = 1
        while (recursive_depth is None or current_depth <= recursive_depth) or len(self.waiting_urls) > 0:
            loop = asyncio.get_event_loop()
            tasks = [self.fetch(url, recursive=(current_depth != recursive_depth)) for url in self.waiting_urls]
            resutls = loop.run_until_complete(asyncio.gather(*tasks))
            self.waiting_urls = {url for url in reduce_seqs(resutls) if url not in self.storage and url not in self.fialed_urls}
            current_depth += 1
        for _ in range(self.retry):
            if len(self.fialed_urls) == 0:
                break
            loop = asyncio.get_event_loop()
            tasks = [self.fetch(url, recursive=False) for url in self.fialed_urls]
            self.fialed_urls = set()
            loop.run_until_complete(asyncio.gather(*tasks))

        if len(self.fialed_urls) > 0:
            print("#### Failed Urls ####")
            for url in self.fialed_urls:
                print(url)

    def load_storage(self, file_name, file_format=FileFormat.JSON):
        self.storage = Saver.load(file_name, file_format)

    def dump_storage(self, file_name, file_format=FileFormat.JSON):
        Saver.dump(self.storage, file_name, file_format)


if __name__ == '__main__':
    spider = Spider("https://docs.oracle.com/javase/8/docs/api/allclasses-noframe.html")
    # start = time.time()
    spider.start_crawl(recursive_depth=2)
    spider.dump_storage("data/jdk8/all-class.html.json")